/* C++ implementation of the budding yeast model
This file was adapted from:
Oguz, C.; Laomettachit, T.; Chen, K. C.; Watson, L. T.; Baumann, W. T.; Tyson, J. J. Optimization and Model Reduction
in the High Dimensional Parameter Space of a Budding Yeast Cell Cycle Model. BMC Syst. Biol. 2013, 7 (1), 53.

Our adaptation of this file allows for interfacing with Python code (instead of Matlab), and adds some extra output
variables required for fitting.
*/
 
 //Model parameters
//gamma=u(1); gammaki=u(2); gammacp=u(3); gammatem=u(4);sig=u(5); signet=u(6); ks_n3=u(7);  Jn3=u(8); Dn3=u(9); kd_n3=u(10);ks_k2=u(11); kd_k2=u(12);kdp_i5=u(13); kdp_i5_14=u(14); kp_i5=u(15); kp_i5_n3=u(16); kp_i5_k2=u(17);kp_i5_n2=u(18); kp_i5_b5=u(19);kdp_bf=u(20); kp_bf_b2=u(21);ks_n2=u(22); ks_n2_bf=u(23); kd_n2=u(24);ks_ki=u(25); ks_ki_swi5=u(26); kd_ki=u(27); kd_kip=u(28);kp_ki_e=u(29); e_ki_n3=u(30); e_ki_k2=u(31); e_ki_n2=u(32); e_ki_b5=u(33); e_ki_b2=u(34); kdp_ki=u(35);kdp_ki_14=u(36);ks_b5=u(37); ks_b5_bf=u(38); kd_b5=u(39);kd_b5_20=u(40);ks_b2=u(41); ks_b2_m1=u(42); kd_b2=u(43); kd_b2_20=u(44); kd_b2_h1=u(45);ks_bud_e=u(46); e_bud_n3=u(47); e_bud_n2=u(48); e_bud_b5=u(49); e_bud_b2=u(50); kd_bud=u(51);ks_spn=u(52); kd_spn=u(53); Jspn=u(54);ks_ori_e=u(55); e_ori_b5=u(56); e_ori_b2=u(57); kd_ori=u(58);ks_swi5=u(59); ks_swi5_m1=u(60); kd_swi5=u(61);ka_swi5_14=u(62); ki_swi5_b2=u(63);ka_m1_b2=u(64); ki_m1=u(65);ks_20=u(66); ks_20_m1=u(67);kd_20=u(68);ka_20=u(69);kd_clb5_20_i=u(70);kd_clb2_20_i=u(71);ki_20_ori=u(72);ka_cp_b2=u(73); ki_cp=u(74);ka_h1=u(75); ka_h1_14=u(76); ki_h1=u(77); ki_h1_e=u(78); e_h1_n3=u(79); e_h1_n2=u(80); e_h1_b5=u(81); e_h1_b2=u(82);kdp_net=u(83); kdp_net_14=u(84); kdp_net_px=u(85); kp_net=u(86); kp_net_b2=u(87); kp_net_en=u(88);kp_net_15=u(89);ka_px=u(90); ki_px=u(91); ki_px_p1=u(92); ks_pds=u(93);  kd_pds=u(94);  kd_pds_20=u(95); ka_15=u(96);  ka_15_14=u(97);  ki_15=u(98); ki_15_b2=u(99);ka_tem=u(100);ka_tem_lo=u(101);ka_tem_p1=u(102);  ki_tem=u(103);  ki_tem_px=u(104);  ks_lo=u(105);   ks_lo_m1=u(106);  kd_lo=u(107);  kd_lo_h1=u(108); ka_lo=u(109);  ka_lo_b2=u(110);  ki_lo=u(111); kas_net=u(112); f=u(113); MDT=u(114); WHI5T=u(115); SBFT=u(116); MCM1T=u(117); APCPT=u(118); CDH1T=u(119); NET1T=u(120); CDC14T=u(121); PPXT=u(122); ESP1T=u(123); CDC15T=u(124); TEM1T=u(125); kd_pds_20_i=u(126);


//Initial conditions also varied during optimization
//V,0=u(127); Cln3,0=u(128); Bck2,0=u(129); WHI5dep,0=u(130); SBFdep,0=u(131);  Cln2,0=u(132);  CKIT,0=u(133);  CKIP,0=u(134);  Clb5T,0=u(135);  Clb2T,0=u(136);  BUD,0=u(137);  ORI,0=u(138);  SPN,0=u(139);  SWI5T,0=u(140);  CDC20T,0=u(141);   CDC20A_APCP,0=u(142);   APCP,0=u(143);      CDH1A,0=u(144);   NET1deP,0=u(145);    PPX,0=u(146); PDS1T,0=u(147);    CDC15,0=u(148);      TEM1,0=u(149);      POLOT,0 =u(150);     POLOA,0=u(151);     CDC20A_APC,0=u(152); 


#include <math.h>
#include <stdio.h>

#if !defined(MAX)
#define	MAX(A, B)	((A) > (B) ? (A) : (B))
#endif

#if !defined(MIN)
#define	MIN(A, B)	((A) < (B) ? (A) : (B))
#endif

//Prototypes
extern void model2(double t, double* y, double* param, double* dydt);  //computes rhs of ode
void eulmod(void (*model)(double, double*, double*,double*), 
        double* tspan, double* y0, double ssize, double* param,
        int nstates, int nsteps, double* tout); //performs euler integration

//Globals 
double CLB2,CLB5,CLB2old,CLB5old,MASS;
double KEZ=0.2, KEZ2=0.2, fDivide=0.4;
// More Globals! Because I want to save this stuff:
double ESP1=0, SBFoverV=0, CLB2CLB5=0, FLAG_UDNA=0, FLAG_BUD=0, FLAG_SPC=0;
double DIV_COUNT = 0;


//Inline functions
inline double H(double totalnum, double sig, double wfunction) {
    return totalnum/(1.+exp(-sig*wfunction));
}
inline double Heav(double x) {
    return (x>0)?1:0;  
}

//Gateway function called by Python
void mexFunction( double tout[],
                  double prhs[],
                  int nY0, double y0[],
                  double tspanValue, double stepSizeValue)
{
/*
Runs the model. This function is to be interfaced with python

Note that the original mxArray input objects count have multiple elements of different sizes. 
I had to change this to multiple separate inputs. 

Inputs
tout[] pointer to 2D array of size (1+output_params) by num_time_steps where to store the output
But actually, passing a 2D array is harder than it looks. Instead, we implement tout as a 1D array with a hard-coded 27 columns. Instead of saying t[variable][time], say t[variable+36*time]
prhs[] array of parameter values to use
y0[] array of initial conditions to use
tspanValue - the final time point of the simulation
stepSizeValue - time step
*/
// Set these all again just to be sure - DIV_COUNT was caught persisting over multiple calls!
// printf("Starting this call with a divcount of %f\n", DIV_COUNT);

ESP1=0; SBFoverV=0; CLB2CLB5=0; FLAG_UDNA=0; FLAG_BUD=0; FLAG_SPC=0; DIV_COUNT = 0;

// printf("Now I reset your divcount to %f\n", DIV_COUNT);

  double *param;
  int nstates,nsteps;
  
  //Get pointers to input data
  double tspan[2] = {0, tspanValue};
  // y0 is set already to the required double array. 
  double stepSize[1] = {stepSizeValue}; 
  param = prhs;
  
  //compute number of states and steps
  nstates=nY0;  
  nsteps=(int)abs((tspan[1]-tspan[0])/stepSize[0])+1;
  if((double)(nsteps-1)*stepSize[0]+tspan[0]<tspan[1])
      nsteps += 1;  //make sure we have enough steps to get to tfinal
      
  // Create matrices for the return arguments. 
  // Actually don't create anything, that should have been created already; we were passed a pointer to it.
  /*plhs[0] = mxCreateDoubleMatrix(nsteps,1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(nsteps,nstates,mxREAL);*/
  
 //Get pointers to output data
 //Nothing doing - tout, our only output structure, is already passed as a pointer. 
  /*tout=&(plhs[0]);//pointer to the start of row 0 of plhs
  yout=&(plhs[1]);//pointer to the start of row 1 of plhs*/
 
 //compute inital values for globals
#define y0(A) y0[(A-1)] // deal with C to Matlab subscript convention - still have to do this because the indices below assume 1-based indexing. 
  CLB5= y0(9) * (y0(9) + y0(10) - y0(7)) / (0.001 + y0(9) + y0(10)) * Heav(y0(9) + y0(10) - y0(7));
  CLB2= y0(10) * (y0(9) + y0(10) - y0(7)) / (0.001 + y0(9) + y0(10)) * Heav(y0(9) + y0(10) - y0(7));
#undef y0 
//Call Euler
  eulmod(&model2,tspan,y0,*stepSize,param,nstates,nsteps,tout);
  
}//End Gateway

//Oak's model


#define dydt(A) dydt[(A-1)]  //deal with C to MATLAB subscript convention
#define y(A) y[(A-1)]
#define u(A) u[(A-1)]       //should be no time penalty since A is const



void model2(double t, double *y, double *u, double* dydt) {

/*global f
global KEZ KEZ2
global CLB5 CLB5old CLB2 CLB2old
global SBF SBFold
global MASS ORI SPN BUD*/


// Constant Parameters 


const double 

//compute the growth rate constant from mass doubling time
mu=log(2.)/u(114);
//fDivide is the fraction of mass retained by the daughter at division
fDivide=u(113);

// Keep old values for event checking 
CLB5old=CLB5;
CLB2old=CLB2;



// State Variables 
double CLN3,BCK2,WHI5deP,SBFdeP,CLN2,CKIT,CKIP,CLB5T,CLB2T,
        BUD,ORI,SPN,SWI5T,CDC20T,APCP,CDH1A,NET1deP,PPX,
	PDS1T,CDC15,TEM1,POLOT,POLOA,APCPT,CDC20A_APCP,CDC20A_APC,CDC20A_APC_T,CDC20A_APCP_T;

MASS=y(1);
CLN3=y(2);
BCK2=y(3);
WHI5deP=y(4);
SBFdeP=y(5);
CLN2=y(6);
CKIT=y(7);
CKIP=y(8);
CLB5T=y(9);
CLB2T=y(10);
BUD=y(11);
ORI=y(12);
SPN=y(13);
SWI5T=y(14);
CDC20T=y(15);
CDC20A_APCP=y(16);
APCP=y(17);
CDH1A=y(18);
NET1deP=y(19);
PPX=y(20);
PDS1T=y(21);
CDC15=y(22);
TEM1=y(23);
POLOT=y(24);
POLOA=y(25);
CDC20A_APC=y(26);




double SBF,CDC14,MEN,W,MCM1,SWI5A;

CLB5= CLB5T * (CLB5T + CLB2T - CKIT) / (0.001 + CLB5T + CLB2T) * Heav(CLB5T + CLB2T - CKIT);

CLB2= CLB2T * (CLB5T + CLB2T - CKIT) / (0.001 + CLB5T + CLB2T) * Heav(CLB5T + CLB2T - CKIT);
CLB2CLB5 = CLB2+CLB5;

SBF= (SBFdeP - WHI5deP) * Heav(SBFdeP - WHI5deP);
SBFoverV = SBF / MASS;

CDC14= (u(121) - u(112)*NET1deP) * Heav(u(121) - u(112)*NET1deP);

ESP1= (u(123) - PDS1T) * Heav(u(123) - PDS1T);

MEN= TEM1 * Heav(CDC15 - TEM1) +CDC15 * Heav(TEM1 - CDC15) - (TEM1) * Heav(CDC15 - TEM1) * Heav(TEM1 - CDC15);

W= u(64) * CLB2 - u(65);
MCM1= H(u(117),u(5),W);

W= u(62) * CDC14 - u(63) * CLB2;
SWI5A= H(SWI5T,u(5),W);


// Differential Equations 

// y(1): MASS
dydt(1)=mu * MASS;

// y(2): CLN3
dydt(2)=u(7) * u(9) * MASS / (u(8) + u(9) * MASS) - u(10) * CLN3;

// y(3): BCK2
dydt(3)=u(11) * MASS - u(12) * BCK2;

//y(4): WHI5deP
W= u(13) + u(14) * CDC14 - u(15) - u(16) * CLN3 - u(17) * BCK2 - u(18) * CLN2 - u(19) * CLB5;
dydt(4)= u(1) * (H(u(115),u(5),W) - WHI5deP);

// y(5): SBFdeP
W= u(20) - u(21) * CLB2;
dydt(5)= u(1) * (H(u(116),u(5),W) - SBFdeP);

// y(6): CLN2
dydt(6)= u(22) + u(23) * SBF - u(24) * CLN2;

// y(7): CKIT
dydt(7)= u(25) + u(26) * SWI5A - u(27) * (CKIT - CKIP) - u(28) * CKIP;

// y(8): CKIP
W= u(29) * (u(30) * CLN3 + u(31) * BCK2 + u(32) * CLN2 + u(33) * CLB5 + u(34) * CLB2) - u(35) - u(36) * CDC14;
dydt(8)= u(2) * (H(CKIT,u(5),W) - CKIP) - u(28) * CKIP;

// y(9): CLB5T
dydt(9)= u(37) + u(38) * SBF - (u(39) + u(40) * CDC20A_APCP + u(70) * CDC20A_APC) * CLB5T;


// y(10): CLB2T
dydt(10)= (u(41) + u(42) * MCM1)*MASS - (u(43) + u(44) * CDC20A_APCP + u(71) * CDC20A_APC + u(45) * CDH1A) * CLB2T;


// y(11): BUD
dydt(11)= u(46) * (u(47) * CLN3 + u(48) * CLN2 + u(49) * CLB5 + u(50) * CLB2) - u(51) * BUD;

// y(12): ORI
dydt(12)= u(55) * (u(56) * CLB5 + u(57) * CLB2) - u(58) * ORI;

// y(13): SPN
dydt(13)= u(52) * Heav(CLB2-u(54)) - u(53) * SPN;

// y(14): SWI5T
dydt(14)= u(59) + u(60) * MCM1 - u(61) * SWI5T;

// y(15): CDC20T
dydt(15)= u(66) + u(67) * MCM1 - u(68) * CDC20T;

// y(16): CDC20A
W= u(69) - u(72) * Heav(ORI - 1.0) * (1.0 - Heav(SPN - 1.0));
if (CDC20T<=APCP)
    CDC20A_APCP_T=CDC20T;
else
    CDC20A_APCP_T=APCP;

dydt(16)= u(1) * (H(CDC20A_APCP_T,u(5),W) - CDC20A_APCP) - u(68) * CDC20A_APCP;

// y(17): APCP
W= u(73) * CLB2 - u(74);
dydt(17)= u(3) * (H(u(118),u(5),W) - APCP);

// y(18): CDH1A
W= u(75) + u(76) * CDC14 - u(77) - u(78) * (u(79) * CLN3 + u(80) * CLN2 + u(81) * CLB5 + u(82) * CLB2);
dydt(18)= u(1) * (H(u(119),u(5),W) - CDH1A);

// y(19): NET1deP
W= u(83) + u(84) * CDC14 + u(85) * PPX - u(86) - u(87) * CLB2 - u(88) * MEN -u(89) * CDC15;
dydt(19)= u(1) * (H(u(120),u(6),W) - NET1deP);

// y(20): PPX
W= u(90) - u(91) - u(92) * ESP1;
dydt(20)= u(1) * (H(u(122),u(5),W) - PPX);

// y(21): PDS1T
dydt(21)= u(93) - (u(94) + u(95) * CDC20A_APCP + u(126) * CDC20A_APC) * PDS1T;


// y(22): CDC15
W= u(96) + u(97) * CDC14 - u(98) - u(99) * CLB2;
dydt(22)= u(1) * (H(u(124),u(5),W) - CDC15);

// y(23): TEM1
W= u(100) + u(101) * POLOA + u(102) * ESP1 - u(103) - u(104) * PPX;
dydt(23)= u(4) * (H(u(125),u(5),W) - TEM1);

// y(24): POLOT
dydt(24)= u(105) + u(106) * MCM1 - (u(107) + u(108) * CDH1A) * POLOT;

// y(25): POLOA
W= u(109) + u(110) * CLB2 - u(111);
dydt(25)= u(1) * (H(POLOT,u(5),W) - POLOA) - (u(107) + u(108) * CDH1A) * POLOA;


// y(26): CDC20A_APC
W= u(69) - u(72) * Heav(ORI - 1.0) * (1.0 - Heav(SPN - 1.0));


if (CDC20T-CDC20A_APCP_T<=u(118)-APCP)
CDC20A_APC_T=CDC20T-CDC20A_APCP_T;
else
CDC20A_APC_T=u(118)-APCP;

dydt(26)= u(1) * (H(CDC20A_APC_T,u(5),W) - CDC20A_APC) - u(68) * CDC20A_APC;


}
#undef dydt   //get rid of macros 
#undef y
#undef u 

//Euler integration

void eulmod(void (*model)(double, double*, double*,double*), 
        double* tspan, double* y0, double ssize, double* param,
        int nstates, int nsteps, double* tout) {
/*function [tout, yout] = eulmod(FunFcn, tspan, y0, ssize,u)

% EUL	Integrates a system of ordinary differential equations using
%	Euler's method.  See also ODE45 and ODEDEMO.
%	[t,y] = eul('yprime', tspan, y0) integrates the system
%	of ordinary differential equations described by the M-file
%	yprime.m over the interval tspan = [t0,tfinal] and using initial
%	conditions y0.
%	[t, y] = eul(F, tspan, y0, ssize) uses step size ssize
%
% INPUT:
% F     - String containing name of user-supplied problem description.
%         Call: yprime = fun(t,y) where F = 'fun'.
%         t      - Time (scalar).
%         y      - Solution vector.
%         yprime - Returned derivative vector; yprime(i) = dy(i)/dt.
% tspan = [t0, tfinal], where t0 is the initial value of t, and tfinal is
%         the final value of t.
% y0    - Initial value vector.
% ssize - The step size to be used. (Default: ssize = (tfinal - t0)/100).
%
% OUTPUT:
% t  - Returned integration time points (column-vector).
% y  - Returned solution, one solution row-vector per tout-value.
% CHANGED TO: tout is a 36 by num_time_steps array, implemented as a 1D array with indexing [i+36*j]
The output has 36 columns (includes auxiliary variables), but the y and dydt vectors have length 26, corresponding
to only the main variables.
Both of these sizes are hard-coded, and the now-ambiguous argument nstates is ignored.
% t_out[0+36*t] gives time at time t
% t_out[i+36*t] gives state i at time t
%*/

// Initialization
// printf("As expected, nstates is %i and nsteps is %i.",nstates,nsteps);

double t0,tfinal,pm=1.,h,dt;
int ind,ind2,k;

//Need to switch these to plain arrays
/*y=(double*)mxCalloc(nstates,sizeof(double));  //make space for vectors
dydt=(double*)mxCalloc(nstates,sizeof(double));*/
double y[26];
double dydt[26];

t0=tspan[0];
tfinal=tspan[1];
if (tfinal<t0) pm = -1.;  // Which way are we computing?
if (ssize < 0) ssize = -ssize; 
h = pm*ssize; // Step size with +or- sign
//t = t0;
//y = y0(:);


// We need to compute the number of steps. (Passed in C version)!

// dt = abs(tfinal - t0);
// N = floor(dt/ssize) + 1;
// if ((N-1)*ssize < dt) N=N+1;

// Initialize the output.

tout[0+36*0] = t0;

for (ind=0,ind2=0;ind2<26;ind += nsteps, ind2++) {
    tout[(ind+1)+36*0]=y0[ind2];
    y[ind2]=y0[ind2];
}
tout[27+36*0] = ESP1;
tout[28+36*0] = SBFoverV;
tout[29+36*0] = CLB2CLB5;
tout[30+36*0] = FLAG_UDNA;
tout[31+36*0] = FLAG_BUD;
tout[32+36*0] = FLAG_SPC;
tout[33+36*0] = DIV_COUNT;
tout[34+36*0] = CLB2;
tout[35+36*0] = CLB5;

k = 1; // Step counter

// The main loop

while (k < nsteps) {
  // Run time step k

  tout[0+36*k]=k*h;
  // Edge case for last time step
  if (pm*(tout[0+36*k] - tfinal) > 0) {
    h = tfinal - tout[0+36*(k-1)];
    tout[0+36*k] = tfinal;
  }
  // Call model() to compute the derivatives of everything. 
  model(tout[0+36*k],y,param,dydt);
  for (ind2=0;ind2<26;ind2++) {
      y[ind2] += h*dydt[ind2];
      if (y[ind2]<0) y[ind2]=0;
  }
// Checking events in cell cycle
// Tip: indices on tout are the same as indices in model() above, because tout is 0-based, but has extra variable time at position 0
// But, indices on y[] are 1 lower, because that's 0-based with no t.

// START transition
// Ugh come back to this one last

// BUD emergence
if (y[10] >= 1 && tout[11+36*(k-1)] < 1 && FLAG_BUD==0) {
    FLAG_BUD = 1;
}

// ORI activation
if (y[11] >= 1 && tout[12+36*(k-1)] < 1 && FLAG_UDNA==0) {
    FLAG_UDNA = 1;
}

// SPN completion
if (y[12] >= 1 && tout[13+36*(k-1)] < 1 && FLAG_SPC==0) {
    FLAG_SPC = 1;
}

// Cell Division 

    if((CLB2 < KEZ) && (CLB2old > KEZ)) {    
        y[0] =fDivide*MASS; // Mass division 
        y[10]=0;         // Reset BUD
        DIV_COUNT += 1;
        FLAG_BUD = 0;
        FLAG_SPC = 0;
        FLAG_UDNA = 0;
    }

    // ORI/SPN 

    if ((CLB2+CLB5 < KEZ2) && (CLB2old+CLB5old> KEZ2)) {
        y[11]=0.0;           // ORI reset
        y[12]=0.0;           // SPN reset
    }
    
    // Step complete, now update output
    //This line makes no sense to me, so I fully rewrote it. 
    //for(ind=k,ind2=0;ind2<nstates;ind+=nsteps,ind2++) yout[ind]=y[ind2];
    for (ind2=0;ind2<26;ind2++) tout[(ind2+1)+36*k]=y[ind2];
    // Then manually record the new auxiliary variables
    tout[27+36*k] = ESP1;
    tout[28+36*k] = SBFoverV;
    tout[29+36*k] = CLB2CLB5;
    tout[30+36*k] = FLAG_UDNA;
    tout[31+36*k] = FLAG_BUD;
    tout[32+36*k] = FLAG_SPC;
    tout[33+36*k] = DIV_COUNT;
    tout[34+36*k] = CLB2;
    tout[35+36*k] = CLB5;
    

    k = k+1;
}
/*mxFree((void*)y);       //free space 
mxFree((void*)dydt);*/
}

/* Example call to see if standalone script works */
int main() {

    double outputArray[36*101];
    double icArray[26];
    for (int i=0; i<26; i++) icArray[i]=1;
    double paramArray[126];
    for (int i=0; i<126; i++) paramArray[i]=0.01;
    paramArray[113]=100.0;
    
    mexFunction(outputArray,
                  paramArray,
                  33, icArray,
                  100.0, 1.0);
    printf("Heres row 17 of your output:\n");
    for (int i=0; i<36; i++) printf("%f,", outputArray[i+36*17]);
    printf("\n");

}
